/*
    Copyright (C) 2005 by Jan Lepper

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include "cssysdef.h"

#include "csgaction.h"
#include "csgplugin.h"
#include "csgmesh.h"

#include "iengine/mesh.h"
#include "iengine/sector.h"
#include "iengine/engine.h"
#include "app/icsstudio.h"
#include "csutil/objreg.h"


SCF_IMPLEMENT_IBASE(cssCsgAction)
SCF_IMPLEMENTS_INTERFACE(icssAction)
SCF_IMPLEMENT_IBASE_END



csPoly3D FlipPolygon(const csPoly3D& poly)
{
  csPoly3D flipped;
  for(size_t v = 1; v <= poly.GetVertexCount(); v++)
    flipped.AddVertex(*poly.GetVertex(poly.GetVertexCount() - v));
  return flipped;
}


bool SplitPolygon(const csPoly3D& poly, const csPoly3D& splitter, csPoly3D& p1, csPoly3D& p2)
{
  
  //TODO: handle coplanar polygons

  csPlane3 SplittingPlane = splitter.ComputePlane();
  csPlane3 PolyPlane = poly.ComputePlane();
  
  //test if the splitter intersects the polygon's plane
  if(splitter.Classify(PolyPlane) != CS_POL_SPLIT_NEEDED)
    return false;

  //test if the splitting plane intersects the polygon
  if(poly.Classify(SplittingPlane) != CS_POL_SPLIT_NEEDED)
    return false;


  poly.SplitWithPlane(p1, p2, SplittingPlane);

  return true;

}


void SplitPolygon(const csPoly3D& polygon, cssPolyArray& splitted_polys, cssCsgMesh& clipper)
{
 
  splitted_polys.Push(polygon);

  for (size_t p = 0; p < clipper.polys.Length(); p++)
  {

     cssPolyArray new_splitted_polys;
     bool clipped = false;

     for(size_t sp = 0; sp < splitted_polys.Length(); sp++)
     {
	  
       csPoly3D p1, p2;
	  
       if(SplitPolygon(splitted_polys[sp], clipper.polys[p].poly, p1, p2))
       {
         new_splitted_polys.Push(p1);
         new_splitted_polys.Push(p2);
         clipped = true;
       }
       else
         new_splitted_polys.Push(splitted_polys[sp]);
     } 

     if (clipped)
       splitted_polys = new_splitted_polys;

  }

}

//-------------------------------------------------------------------------


cssCsgAction::cssCsgAction(iBase *parent)
{
  SCF_CONSTRUCT_IBASE(parent);
  
  csg_operation = CSS_CSG_ADD;
  output_mesh_type = CSS_MESHTYPE_THING;
  remove_source_meshes = true;
}


cssCsgAction::~cssCsgAction()
{
  SCF_DESTRUCT_IBASE();
}


bool cssCsgAction::Initialize(iCStudioCS *csi)
{
  this->csi = csi;

  return true;
}


bool cssCsgAction::Do()
{
    
  csRef<iEngine> engine (CS_QUERY_REGISTRY(csi->getObjectRegistry(), iEngine));

  if(!sector)
    sector = csi->getActiveSector();
  
  if(!sector)
    return false;

  cssCsgMesh m1, m2;
  cssCsgMesh newmesh(engine, sector, "new mesh", output_mesh_type); /// FIXME: prompt for name

  if(!m1.ImportMesh(mesh1) || !m2.ImportMesh(mesh2))
     return false;

  CsgOperation(m1, m2, newmesh);
  
  if(remove_source_meshes)
  {
    // TODO - what to do with the factory(s) ?
    engine->RemoveObject(mesh1);
    engine->RemoveObject(mesh2);
  }
  
  output_mesh = newmesh.Build();
  //engine->Prepare();
  engine->ForceRelight();

  csRef<icssContentCallback> ccb = SCF_QUERY_INTERFACE(csi, icssContentCallback);
  ccb->OnMeshListChanged();
 
  return true;
}


bool cssCsgAction::Undo()
{
  csRef<iEngine> engine (CS_QUERY_REGISTRY(csi->getObjectRegistry(), iEngine));

  engine->RemoveObject(output_mesh);

  if(remove_source_meshes)
  {
    sector->GetMeshes()->Add(mesh1);
    sector->GetMeshes()->Add(mesh2);
  }
  
  //engine->Prepare();
  engine->ForceRelight();

  csRef<icssContentCallback> ccb = SCF_QUERY_INTERFACE(csi, icssContentCallback);
  ccb->OnMeshListChanged();
  
  return true;
}


bool cssCsgAction::Redo()
{
  return Do();
}


void cssCsgAction::SetSourceMeshes(iMeshWrapper* mesh1, iMeshWrapper* mesh2)
{
  this->mesh1 = mesh1;
  this->mesh2 = mesh2;
}


void cssCsgAction::SetOperation(int operation)
{
  csg_operation = operation;
}


void cssCsgAction::CsgOperation(cssCsgMesh& mesh1, cssCsgMesh& mesh2, cssCsgMesh& new_mesh)
{
   cssBspTree bsp1 , bsp2;
   bsp1.Build(mesh1);
   bsp2.Build(mesh2);
   
   ProcessMesh(mesh1, mesh2, new_mesh, bsp1, bsp2, 0);
   ProcessMesh(mesh2, mesh1, new_mesh, bsp1, bsp2, 1);
}


void cssCsgAction::ProcessMesh(
        cssCsgMesh& mesh,
        cssCsgMesh& clipper,
        cssCsgMesh& new_mesh,
        cssBspTree& bsp1,
        cssBspTree& bsp2,
        bool mesh_to_process)
{
  for(size_t p = 0; p < mesh.polys.Length(); p++)
  {
      cssPolyArray clipped;
      SplitPolygon(mesh.polys[p].poly, clipped, clipper);
    
      for(size_t cp = 0; cp < clipped.Length(); cp++)
      {
         csPoly3D poly = clipped[cp];
         if(poly.GetVertexCount() < 3)
                continue;

  
         csVector3 center = poly.GetCenter();
         bool remove = false;
         bool flip = false;
  
         if(csg_operation == CSS_CSG_ADD)
         {
            if((mesh_to_process == 0 && bsp2.PointInside(center)) ||
               (mesh_to_process == 1 && bsp1.PointInside(center)))
              remove = true;
         }
         else if(csg_operation == CSS_CSG_SUBTRACT)
         {
           if(mesh_to_process == 0)
	    {
	       if(bsp2.PointInside(center))
  	       remove = true;
  	     }
	    if(mesh_to_process == 1)
	    {
              if(bsp1.PointInside(center))
	        flip = true;
	      else
	        remove = true;
             }
          }
         else if(csg_operation == CSS_CSG_INTERSECT)
         {
            if((bsp1.Classify(center) < 0) || (bsp2.Classify(center) < 0))
	      remove = true;
         }
  
        if(remove)
          continue;
      
        if(flip)
          poly = FlipPolygon(poly);

        new_mesh.polys.Push(cssCsgPolygon(poly, mesh.polys[p].material));
	
      }
  }
}
