/*
    Copyright (C) 2003 by Philipp Aumayr

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include "cssysdef.h"

#include "wx/wxprec.h"

#ifndef WX_PRECOMP
        #include "wx/wx.h"
#endif


#include "igeom/polymesh.h"
#include "iengine/engine.h"
#include "iengine/mesh.h"
#include "iengine/movable.h"
#include "imesh/object.h"
#include "igeom/objmodel.h"
#include "iengine/sector.h"
#include "iengine/portalcontainer.h"
#include "iengine/portal.h"
#include "csutil/csstring.h"
#include "iutil/objreg.h"
#include "iutil/object.h"

#include "defaultem.h"
#include "app/icsstudio.h"
#include "plugins/icssview.h"
#include "plugins/icssselection.h"
#include "action/icssaddselection.h"
#include "plugins/icssdrawer.h"
#include "ivaria/reporter.h"
#include "sectordialog.h"

CS_IMPLEMENT_PLUGIN


SCF_IMPLEMENT_IBASE(cssDefaultEditModePlugin)
SCF_IMPLEMENTS_INTERFACE(icssPlugin)
SCF_IMPLEMENTS_EMBEDDED_INTERFACE(icssEditMode)
SCF_IMPLEMENT_IBASE_END

SCF_IMPLEMENT_EMBEDDED_IBASE(cssDefaultEditModePlugin::cssEditMode)
SCF_IMPLEMENTS_INTERFACE(icssEditMode)
SCF_IMPLEMENT_EMBEDDED_IBASE_END

SCF_IMPLEMENT_FACTORY (cssDefaultEditModePlugin)

cssDefaultEditModePlugin::cssDefaultEditModePlugin(iBase *parent)
{
  SCF_CONSTRUCT_IBASE(parent);
  SCF_CONSTRUCT_EMBEDDED_IBASE(scficssEditMode);
}

cssDefaultEditModePlugin::~cssDefaultEditModePlugin()
{
  SCF_DESTRUCT_EMBEDDED_IBASE(scficssEditMode);
  SCF_DESTRUCT_IBASE();
}


//// initialisation 
bool cssDefaultEditModePlugin::Initialize(iCStudioCS *csi)
{
  m_csi = csi;
  return true;
}

void cssDefaultEditModePlugin::GetActionManager()
{
  csRef<icssPlugin> amplug = m_csi->GetPlugin("cstudio.plugin.actionmgr");
  m_actionmgr = SCF_QUERY_INTERFACE(amplug, icssActionManager);
}

void cssDefaultEditModePlugin::GetMeshSelector()
{
  csRef<icssPlugin> selectorplug = m_csi->GetPlugin("cstudio.select.object.mesh");
  if(selectorplug == 0) 
  {
    csReport(m_csi->getObjectRegistry(), CS_REPORTER_SEVERITY_ERROR,
		    "crystalspace.application.cstudio",
		    "Error getting cstudio.select.object.mesh plugin");
    return;
  }  else {
    m_meshselector = SCF_QUERY_INTERFACE(selectorplug, icssSelector);
  }
}
  
/*
* icssEditMode implementation
*/
const char *cssDefaultEditModePlugin::GetDescription()
{
  return "Standard";
}

void cssDefaultEditModePlugin::OnActivate(icssEditMode *lastactive)
{
  GetActionManager();
  GetMeshSelector();
}

void cssDefaultEditModePlugin::OnDeactivate(icssEditMode *nextactive)
{
}

/// Event catcher, dynamic_cast to check type
bool cssDefaultEditModePlugin::OnEvent(wxEvent &event, icssView *view)
{
  /*
  # wxEVT_LEFT_DOWN
  # wxEVT_LEFT_UP
  # wxEVT_LEFT_DCLICK
  # wxEVT_MIDDLE_DOWN
  # wxEVT_MIDDLE_UP
  # wxEVT_MIDDLE_DCLICK
  # wxEVT_RIGHT_DOWN
  # wxEVT_RIGHT_UP
  # wxEVT_RIGHT_DCLICK
  # wxEVT_MOTION
  # wxEVT_MOUSEWHEEL 
  */
  if(event.GetEventType() == wxEVT_LEFT_DOWN)
  {
    wxMouseEvent *me = wxDynamicCast(&event, wxMouseEvent);
    if(me)
      return OnLeftDown(me, view);
  }
  if(event.GetEventType() == wxEVT_KEY_DOWN)
  {
    wxKeyEvent *kev = wxDynamicCast(&event, wxKeyEvent);
    
    switch(kev->GetKeyCode())
    {
    case WXK_F3:
      SwitchLocalSector(*kev);
      return true;
    case WXK_ESCAPE:
      ClearSelection();
      return true;
    }
  }
  // if we don't do anything we have to skip the event
  event.Skip();
  return false;
}

void cssDefaultEditModePlugin::SwitchLocalSector(wxKeyEvent &event)
{
  /// opens a little dialog with a list of adjacent sectors
  /// of the current active sector

  wxPoint mouse(event.m_x, event.m_y);
  wxWindow * wnd = wxDynamicCast(event.GetEventObject(), wxWindow);
  if(wnd)
    mouse = wnd->ClientToScreen(mouse);

  cssSectorDialog dlg(m_csi, NULL, -1, "Select sector", 
    mouse, wxDefaultSize);
  dlg.FillSectors(m_csi->getActiveSector());
  if(dlg.ShowModal() == wxID_OK)
  {
    SetActiveSector((char *)dlg.m_selected_sector.GetData(), true);
    wnd->Refresh();
  }
}

void cssDefaultEditModePlugin::SetActiveSector(csString newsector, bool inall)
{
  // inall decides wether to search in current active sector only
  // or if it's one of all sectors

  csRef<iSector> curactive = m_csi->getActiveSector();
  if(inall == false && curactive.IsValid())
  {
    csSet<csPtrKey<iMeshWrapper> >::GlobalIterator it(curactive->GetPortalMeshes().GetIterator());
    
    while(it.HasNext())
    {
      iMeshWrapper* mwrap = it.Next();
      csRef<iPortalContainer> pc = mwrap->GetPortalContainer();
      
      int nportals = pc->GetPortalCount();

      for(int i = 0; i < nportals; i++)
      {
        csRef<iPortal> p = pc->GetPortal(i);

        csRef<iSector> s = p->GetSector();
      
        if(s.IsValid() && strcmp(s->QueryObject()->GetName(), newsector) == 0)
        {
          m_csi->SetActiveSector(s);
          return;
        }
      }
    }
  }
  else
  {
    csRef<iEngine> ie = CS_QUERY_REGISTRY(m_csi->getObjectRegistry(), iEngine);
    if(!ie)
      return;

    csRef<iSectorList> sl = ie->GetSectors();
    if(sl)
    {
      int count = sl->GetCount();
      for (int i = 0; i < count; i++)
      {
        csRef<iSector> s = sl->Get(i);
        if(s)
        {
          if(strcmp(s->QueryObject()->GetName(), newsector.GetData()) == 0)
            m_csi->SetActiveSector(s);
        }
      }
    }
  }
}

bool cssDefaultEditModePlugin::OnLeftDown(wxMouseEvent *e, icssView *view)
{
  csSegment3 ray = view->GetSegmentAt(e->m_x, e->m_y);

  csRefArray<iBase> newsel;

  // fill newsel with results from the meshselector
  newsel = m_meshselector->HitBeam(ray.Start(), ray.End(), 0.0, newsel);
  /// 
  /// if shift is not pressed, we have to select and not add
  /// -> clear the selection first
	
  if(!e->m_shiftDown)
  {
    csRef<icssAction> clearaction = 
                   SCF_CREATE_INSTANCE("cstudio.action.clearselection", icssAction);
    clearaction->Initialize(m_csi);
    m_actionmgr->Perform(clearaction);
		
  }


  csRef<icssAction> addaction =
                   SCF_CREATE_INSTANCE("cstudio.action.addselection", icssAction);
  addaction->Initialize(m_csi);
  csRef<icssAddSelectionAction> iaddsel = 
                   SCF_QUERY_INTERFACE(addaction, icssAddSelectionAction);

  for(size_t i = 0; i < newsel.Length(); i++)
    iaddsel->Add(newsel.Get(i));
	
  m_actionmgr->Perform(addaction);
  view->Redraw();
  return true;
}


void cssDefaultEditModePlugin::ClearSelection()
{
  csRef<icssAction> clearaction 
                = SCF_CREATE_INSTANCE("cstudio.action.clearselection", icssAction);
  clearaction->Initialize(m_csi);
  m_actionmgr->Perform(clearaction);
}

void cssDefaultEditModePlugin::OnDraw (icssDrawer *d, icssView *v)
{
  csRef<iSector> as = m_csi->getActiveSector();
  if(!as)
    return;

  csColor4 c(0.7f, 0.7f, 0.7f);
  d->SetColor(c);

  
  csRef<iMeshList> ml = as->GetMeshes();
  int mesh_count = ml->GetCount();
  for (int i = 0; i < mesh_count; i++)
  {
    d->DrawMesh(ml->Get(i));
  }
}
