/*
    Copyright (C) 2005 by Jan Lepper

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include "cssysdef.h"
#include "csutil/ref.h"
#include "csgeom/csrect.h"
#include "app/icsstudio.h"
#include "plugins/icssplugin.h"
#include "plugins/icsscreateview.h"
#include "splitview4wnd.h"
#include "iutil/document.h"
#include "iutil/stringarray.h"


#define CSS_SPLITTER_BORDER 2

BEGIN_EVENT_TABLE (cssSplitView4Wnd, wxWindow)
  EVT_MOUSE_EVENTS(cssSplitView4Wnd::OnMouse)
  EVT_PAINT(cssSplitView4Wnd::OnPaint)
  EVT_ERASE_BACKGROUND(cssSplitView4Wnd::OnErase)
  EVT_MOUSE_CAPTURE_CHANGED(cssSplitView4Wnd::OnCaptureChanged)
  EVT_SIZE(cssSplitView4Wnd::OnSize)
  EVT_MENU_RANGE(CSS_VIEW1_START, CSS_VIEW_END, cssSplitView4Wnd::OnMenu)
END_EVENT_TABLE()


cssSplitView4Wnd::cssSplitView4Wnd(wxWindow *parent, iCStudioCS *csi, wxPoint pos, wxSize size) : 
  wxWindow(parent, -1, pos, size) 
{
  this->csi = csi;
  
  for(int i = 0; i < 4; i++)
    views[i].wnd = 0;

  update = false;
  dragging_x = dragging_y = false;  
  
  GetClientSize(&clientsize.x, &clientsize.y);

  splitter_pos.Set(0.5, 0.5);
  GetSplitterPosAbs();
  
  menu = new wxMenu();
  menu->Append(0, "top/left", CreatePluginMenu(CSS_VIEW1_START));
  menu->Append(0, "top/right", CreatePluginMenu(CSS_VIEW2_START));
  menu->Append(0, "bottom/left", CreatePluginMenu(CSS_VIEW3_START));
  menu->Append(0, "bottom/right", CreatePluginMenu(CSS_VIEW4_START));
  

  for(int i = 0; i < 4; i++)
    SetView(i, "cstudio.view.view2d");  
  
  Layout();
  
}


void cssSplitView4Wnd::SetView(int view, const char* plugin)
{
  View *v = &views[view];
  
  if(v->wnd)
    v->wnd->Destroy();
  v->view = 0;
  
  csRef<icssPlugin> plg = csi->GetPlugin(plugin);
  
  v->view = SCF_QUERY_INTERFACE(plg, icssCreateView);
  v->wnd = v->view->Create(this, "", 0,0,0,0);
}


void cssSplitView4Wnd::SetView(int view, int plugin)
{
  csRef<iStringArray> vp = iSCF::SCF->QueryClassList("cstudio.view.");
  vp->Sort();
  SetView(view, vp->Get(plugin));
}


wxMenu *cssSplitView4Wnd::CreatePluginMenu(int startid)
{
  wxMenu *m = new wxMenu();

  csRef<iStringArray> vp = iSCF::SCF->QueryClassList("cstudio.view.");
  vp->Sort();

  size_t count = vp->Length();

  for(size_t i = 0; i < count; i++)
    m->Append(startid + i, vp->Get(i));

  return m;
}


bool cssSplitView4Wnd::Load(iDocumentNode *parent)
{
  splitter_pos.x = parent->GetAttributeValueAsFloat("splitter_x");
  splitter_pos.y = parent->GetAttributeValueAsFloat("splitter_y");

  csRef<iDocumentNodeIterator> it = parent->GetNodes("view");
  
  
  for(int i = 0; i < 4 && it->HasNext(); i++)
  {
    csRef<iDocumentNode> viewnode = it->Next();
    const char* classid = viewnode->GetAttributeValue("classid");
    
    if(!classid)
      classid = "cstudio.view.view2d";
    
    SetView(i, classid);
    views[i].view->LoadXML(views[i].wnd, viewnode);
  }  

  Layout();
  return true;
}


bool cssSplitView4Wnd::Save(iDocumentNode *parent)
{
  parent->SetAttributeAsFloat("splitter_x", splitter_pos.x);
  parent->SetAttributeAsFloat("splitter_y", splitter_pos.y);
  
  for(int i = 0; i < 4; i++)
  {
    csRef<icssPlugin> plg = SCF_QUERY_INTERFACE(views[i].view, icssPlugin);
    csRef<iDocumentNode> viewnode = parent->CreateNodeBefore(CS_NODE_ELEMENT);
    viewnode->SetValue("view");
    viewnode->SetAttribute("classid", plg->GetName());
    views[i].view->SaveXML(views[i].wnd, viewnode);
  }
  
  return true;
}


void cssSplitView4Wnd::GetSplitterPosAbs()
{
  float abs_x = splitter_pos.x * (float)clientsize.x;
  float abs_y = splitter_pos.y * (float)clientsize.y;
   
  if(abs_x > clientsize.x - 4)
    abs_x = clientsize.x - 4;
  if(abs_x < 3)
    abs_x = 3;
  if(abs_y > clientsize.y - 4)
    abs_y = clientsize.y - 4;
  if(abs_y < 3)
    abs_y = 3;
     
  splitter_pos_abs.Set(abs_x, abs_y);
}  


void cssSplitView4Wnd::SetSplitterPosAbs(const csVector2 &p)
{
  splitter_pos.x = p.x / (float)clientsize.x;
  splitter_pos.y = p.y / (float)clientsize.y;

  if(p.x < 7)
    splitter_pos.x = 0;
  if(p.x > clientsize.x - 7)
    splitter_pos.x = 1;

  if(p.y < 7)
     splitter_pos.y = 0;
  if(p.y > clientsize.y - 7)
     splitter_pos.y = 1;
       
  GetSplitterPosAbs();
  Layout();
}  

void cssSplitView4Wnd::OverSplitter(const csVector2 &p, bool *over_x, bool *over_y) 
{ 
  *over_x = p.x >= splitter_pos_abs.x - 2 && 
       p.x <= splitter_pos_abs.x + 2 && p.y > 0 && p.y < clientsize.y;
  *over_y = p.y >= splitter_pos_abs.y - 2 && 
       p.y <= splitter_pos_abs.y + 2 && p.x >= 0 && p.x <= clientsize.x;
}  


void cssSplitView4Wnd::InvertDraggingPos()
{
  wxClientDC dc(this);
  dc.SetPen(wxPen(*wxBLACK, 3, wxSOLID));
  dc.SetLogicalFunction(wxINVERT);

  csVector2 pt = splitter_pos_dragging;
  
  if(dragging_x)
    dc.DrawLine((int)pt.x, 0, (int)pt.x, clientsize.y);

  if(dragging_y)
    dc.DrawLine(0, (int)pt.y, clientsize.x, (int)pt.y);
}


void cssSplitView4Wnd::EndDragging()
{
  if(dragging_x || dragging_y)
  {
    if(!update)
    {
      InvertDraggingPos();
      SetSplitterPosAbs(splitter_pos_dragging);
    }
    dragging_x = dragging_y = false;
  }
}


void cssSplitView4Wnd::LayoutView(View *view, csRect &r)
{
  view->wnd->SetSize(r.xmin, r.ymin + 0, r.Width(), r.Height() - 0);
}


bool cssSplitView4Wnd::Layout()
{
  csVector2 pos(splitter_pos_abs.x, splitter_pos_abs.y);
  
  //top-left
  csRect r (0, 0, (int)(pos.x - 0 - CSS_SPLITTER_BORDER), (int)(pos.y - 0 - CSS_SPLITTER_BORDER));
  LayoutView(&views[0], r);
  
  //top-right
  r.Set((int)(pos.x + 1 + CSS_SPLITTER_BORDER), 0, clientsize.x, (int)(pos.y - 0 - CSS_SPLITTER_BORDER));
  LayoutView(&views[1], r);  
  
  //bottom-left
  r.Set(0, (int)(pos.y + 1 + CSS_SPLITTER_BORDER), (int)(pos.x - 0 - CSS_SPLITTER_BORDER), clientsize.y);
  LayoutView(&views[2], r);
  
  //bottom-right
  r.Set((int)(pos.x + 1 + CSS_SPLITTER_BORDER), (int)(pos.y + 1 + CSS_SPLITTER_BORDER), clientsize.x, clientsize.y);
  LayoutView(&views[3], r);
  
  Refresh();
  return true;
}


void cssSplitView4Wnd::OnMenu(wxCommandEvent &event)
{
  int id = event.GetId();
  
  if(id < CSS_VIEW2_START)
    SetView(0, id - CSS_VIEW1_START);
  else if(id < CSS_VIEW3_START)
    SetView(1, id - CSS_VIEW2_START);  
  else if(id < CSS_VIEW4_START)
    SetView(2, id - CSS_VIEW3_START);  
  else
    SetView(3, id - CSS_VIEW4_START);  
    
  Layout();
}


void cssSplitView4Wnd::OnMouse(wxMouseEvent &event)
{
  bool over_x, over_y;
  OverSplitter(csVector2(event.m_x, event.m_y), &over_x, &over_y);
  bool over_splitter = over_x || over_y;

  if(event.RightDown())
  {
    if(!dragging_x && !dragging_y)
    {
      SetCursor(*wxSTANDARD_CURSOR);    
      PopupMenu(menu, event.m_x, event.m_y);
    }
    return;
  }
  
  else if(event.LeftDown())
  {
    if(over_splitter)
    {
      splitter_pos_dragging = splitter_pos_abs;
      
      if (over_x)
        dragging_x = true;
      if(over_y)
        dragging_y = true;

      if(over_splitter && GetCapture() != this)
        CaptureMouse();
      if(!update)
        InvertDraggingPos();
    }       
  }
  
  else if(event.LeftUp())
  {
    if(GetCapture() == this && !over_splitter)
      ReleaseMouse();
    EndDragging();      
  }

  else if(event.Moving())
  {
    
    if(dragging_x || dragging_y)
    {
      if(!update)
        InvertDraggingPos();
    
      if(dragging_x)
        splitter_pos_dragging.x = event.m_x;
      if(dragging_y)
        splitter_pos_dragging.y = event.m_y;
      
      if(update)
        SetSplitterPosAbs(splitter_pos_dragging);
      else
        InvertDraggingPos();
    }
    else
    {
      if(over_splitter && GetCapture() != this)
        CaptureMouse();
      else if(!over_splitter && GetCapture() == this)
        ReleaseMouse();
    
      if(over_x && over_y)
        SetCursor(wxCursor(wxCURSOR_SIZING));
      else if(over_x)
        SetCursor(wxCursor(wxCURSOR_SIZEWE));
      else if(over_y)
        SetCursor(wxCursor(wxCURSOR_SIZENS));
      else
        SetCursor(*wxSTANDARD_CURSOR);
    }

  }

  event.Skip();
}


void cssSplitView4Wnd::OnCaptureChanged(wxMouseCaptureChangedEvent &event)
{
  EndDragging();
}


void cssSplitView4Wnd::OnSize(wxSizeEvent &event)
{
  GetClientSize(&clientsize.x, &clientsize.y);
  GetSplitterPosAbs();

  if(views[0].wnd)
    Layout();
}


void cssSplitView4Wnd::OnErase(wxEraseEvent &event)
{
}


void cssSplitView4Wnd::OnPaint(wxPaintEvent &event)
{
  wxPaintDC dc(this);
  csVector2 pos((int)splitter_pos_abs.x, (int)splitter_pos_abs.y);
  
  wxPen pen_splitter (wxSystemSettings::GetColour(wxSYS_COLOUR_3DFACE), 3, wxSOLID);
  wxPen pen_shadow (wxSystemSettings::GetColour(wxSYS_COLOUR_3DSHADOW), 1, wxSOLID);
  wxPen pen_highlight (wxSystemSettings::GetColour(wxSYS_COLOUR_3DHIGHLIGHT), 1, wxSOLID);

  //left and upper border  
  dc.SetPen(pen_highlight);
  dc.DrawLine(0, (int)(pos.y-2), clientsize.x, (int)(pos.y-2)); // y-splitter
  dc.DrawLine((int)(pos.x-2), 0, (int)(pos.x-2), clientsize.y); // x-splitter
  
  //right and lower border
  dc.SetPen(pen_shadow);
  dc.DrawLine(0, (int)(pos.y+2), clientsize.x, (int)(pos.y+2)); // y-splitter
  dc.DrawLine((int)(pos.x+2), 0, (int)(pos.x+2), clientsize.y); // x-splitter
  
  //splitter faces
  dc.SetPen(pen_splitter);
  dc.DrawLine(0, (int)pos.y, clientsize.x, (int)pos.y); // y-splitter
  dc.DrawLine((int)pos.x, 0, (int)pos.x, clientsize.y); // x-splitter
  
  dc.SetPen(wxNullPen);
  dc.SetBrush(wxNullBrush);
}
