#include "wx/glcanvas.h"

#include "cssysdef.h"
#include "igeom/polymesh.h"
#include "iengine/engine.h"
#include "iengine/mesh.h"
#include "iengine/movable.h"
#include "imesh/object.h"
#include "igeom/objmodel.h"
#include "iengine/scenenode.h"
#include "iengine/sector.h"
#include "iengine/portalcontainer.h"
#include "iengine/portal.h"
#include "iengine/light.h"

#include "csutil/cscolor.h"

#include "plugins/icssselection.h"

#include "drawer.h"
 
SCF_IMPLEMENT_IBASE(cssDrawer)
SCF_IMPLEMENTS_INTERFACE(icssDrawer)
SCF_IMPLEMENT_IBASE_END

cssDrawer::cssDrawer ()
{
  
  SCF_CONSTRUCT_IBASE(0);
  
  // bottom rect
  bbox_indices[0] = 0;
  bbox_indices[1] = 1;
  bbox_indices[2] = 1;
  bbox_indices[3] = 2;
  bbox_indices[4] = 2;
  bbox_indices[5] = 3;
  bbox_indices[6] = 3;
  bbox_indices[7] = 0;
  
  // top rect
  bbox_indices[8] = 4;
  bbox_indices[9] = 5;
  bbox_indices[10] = 5;
  bbox_indices[11] = 6;
  bbox_indices[12] = 6;
  bbox_indices[13] = 7;
  bbox_indices[14] = 7;
  bbox_indices[15] = 4;

  // connecting lines
  bbox_indices[16] = 0;
  bbox_indices[17] = 4;
  bbox_indices[18] = 1;
  bbox_indices[19] = 5;
  bbox_indices[20] = 2;
  bbox_indices[21] = 6;
  bbox_indices[22] = 3;
  bbox_indices[23] = 7;

}

cssDrawer::~cssDrawer ()
{
  SCF_DESTRUCT_IBASE();
}
 

void cssDrawer::Initialize(iCStudioCS *csi)
{
  m_csi = csi;
}

void cssDrawer::SetColor(csColor4 &c)
{
  glColor4f(c.red, c.green, c.blue, c.alpha);
  draw_color = c;
    

}

csColor4 &cssDrawer::GetColor()
{
  return draw_color;
}

void cssDrawer::SetLineStipple(int factor, unsigned short pattern)
{
  if(factor)
  {
    glEnable(GL_LINE_STIPPLE);
    glLineStipple(factor, pattern);
    stipple_factor = factor;
    stipple_pattern = pattern;
  }
  else
  {
    glDisable(GL_LINE_STIPPLE);
    stipple_factor = factor;
  }
}

void cssDrawer::GetLineStipple(int &ofactor, unsigned short &opattern)
{
  ofactor = stipple_factor;
  opattern = stipple_pattern;
}

void cssDrawer::SetLineWidth (float width) 
{
  glLineWidth (width);
  line_width = width;
}

float cssDrawer::GetLineWidth () 
{
  return line_width;
}

void cssDrawer::SetPointSize (float size)
{
  glPointSize(size);
  point_size = size;
}

float cssDrawer::GetPointSize ()
{
  return point_size;
}
 
void cssDrawer::DrawMesh (iMeshWrapper *m)
{
   csRef<iMeshObject> mo = m->GetMeshObject();
  
  if(mo.IsValid())
  {
    csRef<iObjectModel> om = mo->GetObjectModel();
    if(om.IsValid())
    {
      csRef<iPolygonMesh> pm = om->GetPolygonMeshBase();
      if(pm.IsValid())
      {
        glPushMatrix();
        ApplyTransform(m->GetMovable()->GetFullTransform());
        DrawPolyMesh(pm);

        glPopMatrix();

        csRefArray<iSceneNode> children = m->QuerySceneNode()->GetChildren();
        for(size_t i = 0; i < children.GetSize(); i++)
          DrawMesh(children.Get(i)->QueryMesh());
      }
    }
  }
}

void cssDrawer::DrawLight(iLight *light)
{
  csVector3 center = light->GetCenter();
  
  csVector3 min = center - csVector3(0.1, 0.1, 0.1);
  csVector3 max = center + csVector3(0.1, 0.1, 0.1);
  
  DrawAABB(min, max);
}

void cssDrawer::DrawObject(iBase *object)
{
  csRef<iMeshWrapper> mesh = SCF_QUERY_INTERFACE(object, iMeshWrapper);
  if(mesh)
  {
    DrawMesh(mesh);
    return;
  }
  
  csRef<iLight> light = SCF_QUERY_INTERFACE(object, iLight);
  if(light)
    DrawLight(light);
}

void cssDrawer::DrawPolyMesh(iPolygonMesh *pm)
{
  if(!pm)
    return;

  
  glEnableClientState(GL_VERTEX_ARRAY);
  
  // Set the vertex pointer
  glVertexPointer(3, GL_FLOAT, 0, pm->GetVertices());
  
  // Draw Polygons
  int pc = pm->GetPolygonCount();
  csMeshedPolygon *polys = pm->GetPolygons();
  for(int i = 0; i < pc; i++)
  {
    glDrawElements(GL_POLYGON, polys[i].num_vertices,GL_UNSIGNED_INT, polys[i].vertices); 
  }

  // Draw Triangles
  
  if(pm->GetTriangleCount() > 0)
    glDrawElements(GL_TRIANGLES, pm->GetTriangleCount() * 3, GL_INT, pm->GetTriangles());


  glDisableClientState(GL_VERTEX_ARRAY);
}

void cssDrawer::DrawAABB (csVector3 &min, csVector3 &max)
{
  LoadBBoxVerts(min, max);

  glEnableClientState (GL_VERTEX_ARRAY);
  
  glVertexPointer(3, GL_FLOAT, 0, bbox_verts);

  glDrawElements (GL_LINES, 24, GL_UNSIGNED_SHORT, bbox_indices);
  
  glDisableClientState (GL_VERTEX_ARRAY);
}

void cssDrawer::LoadBBoxVerts(csVector3 &min, csVector3 &max)
{
  bbox_verts[0] = min;
  
  bbox_verts[1].x = max.x;
  bbox_verts[1].y = min.y;
  bbox_verts[1].z = min.z;

  bbox_verts[2].x = max.x;
  bbox_verts[2].y = min.y;
  bbox_verts[2].z = max.z;

  bbox_verts[3].x = min.x;
  bbox_verts[3].y = min.y;
  bbox_verts[3].z = max.z;

  bbox_verts[4].x = min.x;
  bbox_verts[4].y = max.y;
  bbox_verts[4].z = min.z;

  bbox_verts[5].x = max.x;
  bbox_verts[5].y = max.y;
  bbox_verts[5].z = min.z;

  bbox_verts[6] = max;

  bbox_verts[7].x = min.x;
  bbox_verts[7].y = max.y;
  bbox_verts[7].z = max.z;
}

void cssDrawer::DrawPoint (csVector3 &p)
{
  glBegin(GL_POINTS);
  glVertex3f(p.x, p.y, p.z);
  glEnd();
}

void cssDrawer::DrawCircle (csVector3 &m, float r, int segcount)
{
  if(segcount > 360);
    segcount = 360;

  CalculateCircleVerts(segcount);

  
    
  glEnableClientState(GL_VERTEX_ARRAY);
  glVertexPointer(2, GL_FLOAT, 0, circle_verts);

  // normal x, y circle view
  
  glPushMatrix();
  
    glScalef(r, r, r);
    glTranslatef(m.x, m.y, m.z);
    
    glDrawArrays(GL_LINE_LOOP, 0, segcount);
  glPopMatrix();

  //  x, z view
  glPushMatrix();
    
    glScalef(r, r, r);
    glRotatef(90, 1.0, 0.0, 0.0);

    glTranslatef(m.x, m.y, m.z);

    glDrawArrays(GL_LINE_LOOP, 0, segcount);
  glPopMatrix();

  // z, y view
  glPushMatrix();

    glScalef(r, r, r);
    glRotatef(90, 0.0, 1.0, 0.0);

    glTranslatef(m.x, m.y, m.z);

    glDrawArrays(GL_LINE_LOOP, 0, segcount);

  glDisableClientState(GL_VERTEX_ARRAY);

}

void cssDrawer::CalculateCircleVerts(int segcount)
{
  float seg = 2 * PI / segcount;
  int j = 0;
  for (float i = 0; i < 2 * PI; i += seg, j++)
  {
    circle_verts[j].x = sin(i);
    circle_verts[j].y = cos(i);
  }
}

void cssDrawer::DrawLine (csVector3 &s, csVector3 &e)
{
  glBegin(GL_LINES);
    glVertex3f(s.x, s.y, s.z);
    glVertex3f(e.x, e.y, e.z);
  glEnd();
}

void cssDrawer::ApplyTransform(const csReversibleTransform& t)
{
  GLfloat matrix[16];  
  
  const csMatrix3 &orientation = t.GetO2T();
  const csVector3 &translation = t.GetO2TTranslation();

  matrix[0] = orientation.m11;
  matrix[1] = orientation.m12;
  matrix[2] = orientation.m13;
  matrix[3] = 0.0f;

  matrix[4] = orientation.m21;
  matrix[5] = orientation.m22;
  matrix[6] = orientation.m23;
  matrix[7] = 0.0f;

  matrix[8] = orientation.m31;
  matrix[9] = orientation.m32;
  matrix[10] = orientation.m33;
  matrix[11] = 0.0f;

  matrix[12] = translation.x;
  matrix[13] = translation.y;
  matrix[14] = translation.z;
  matrix[15] = 1.0f;  

  glMultMatrixf (matrix);
}

