/*
    Copyright (C) 2003 by Philipp Aumayr

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "wx/glcanvas.h"
#include "cssysdef.h"
#include "csutil/objreg.h"
#include "csgeom/math3d.h"
#include "igeom/polymesh.h"
#include "iengine/engine.h"
#include "iengine/camera.h"
#include "iengine/mesh.h"
#include "iengine/movable.h"
#include "imesh/object.h"
#include "igeom/objmodel.h"
#include "iengine/sector.h"
#include "iengine/portalcontainer.h"
#include "iengine/portal.h"


#include "app/icsstudio.h"
#include "plugins/icssplugin.h"
#include "plugins/icssselection.h"
#include "plugins/icsseditmode.h"
#include "app/icsseditmodemanager.h"

#include "view2ddrawer.h"
#include "view2dwnd.h"


cssView2DDrawer::cssView2DDrawer(csRect rect, int vt)
{
  viewtype = vt;
  viewarea = rect;
  last_pos.x = last_pos.y = 0.0f;
  pos.x = pos.y = 0.0f;
  zoom = 6.0f;
  fov = 90.0f;
  object_view = false;
}

cssView2DDrawer::~cssView2DDrawer()
{
}

bool cssView2DDrawer::Initialize(iCStudioCS *csi, cssView2DWnd *wnd)
{
  m_csi = csi;
  m_wnd = wnd;
  
  ref_plane = CSS_PLANE_Y;
  
  csRef<iEngine> engine = CS_QUERY_REGISTRY(csi->getObjectRegistry(), iEngine);
  
  camera3d = engine->CreateCamera();
  camera3d->SetFOVAngle(fov, viewarea.Width());
  camera3d->SetPerspectiveCenter(viewarea.Width() / 2, viewarea.Height() / 2);  
  
  return true;
}

void cssView2DDrawer::SetViewArea(int width, int height)
{
  viewarea.Set(0, 0, width, height);
  camera3d->SetFOVAngle(fov, width);
  camera3d->SetPerspectiveCenter(width / 2, height / 2);
}

void cssView2DDrawer::SetFOV(float angle)
{
  fov = angle;
  if(fov < 10)
    fov = 10;
  else if(angle > 140)
    fov = 140;
  camera3d->SetFOVAngle(fov, viewarea.Width());
  camera3d->SetPerspectiveCenter(viewarea.Width() / 2, viewarea.Height() / 2);    
}

void cssView2DDrawer::Draw()
{ 
  csRect maparea;

  float dist = viewarea.Width( ) / 2.0f / zoom; // distance from center
  float mapxmin, mapxmax, mapymin, mapymax;
  mapxmin = pos.x - dist;
  mapxmax = pos.x + dist;  
  dist = viewarea.Height( ) / 2.0f / zoom;
  mapymin = pos.y + dist;  
  mapymax = pos.y - dist;  

  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  glViewport(viewarea.xmin, viewarea.ymin, viewarea.Width(), viewarea.Height());

  
  // setup projection
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  
  if(viewtype == V2D_PERSPECTIVE)
  {
    float aspect = (float)viewarea.Height() / (float)viewarea.Width(); 
    
    float right = tan((camera3d->GetFOVAngle() * PI)  / 360.0) * SMALL_Z;
    float left = -right;
    float top = right * aspect;
    float bottom = left * aspect;
    
    glFrustum(left, right, bottom, top, SMALL_Z, 10000.f);
  }    
  else
    glOrtho(mapxmin, mapxmax, mapymax, mapymin, 1000000, -1000000);


  // setup modelview / draw grid
  
  glMatrixMode(GL_MODELVIEW);  
  glLoadIdentity();  

  glScalef(1, 1, -1);
  
  if(viewtype == V2D_PERSPECTIVE)
  {
    drawer.ApplyTransform(camera3d->GetTransform().GetInverse());
    

    glPushMatrix();
    
    glRotatef(90, 1.0, 0.0, 0.0);

    csVector3 pos = camera3d->GetTransform().GetOrigin();

    grid.GenerateGridLines(pos.x - 1000, pos.x +1000, 10.0, pos.z - 1000, pos.z + 1000, 10.0);
    
    grid.DrawGrid(csColor(0.0f, 0.0f, 0.4f));
    grid.DrawOrigin(csColor(0.0f, 0.0f, 0.6f));    
    
    glPopMatrix();
  }
  else
  {    
    grid.GenerateGridLines(mapxmin, mapxmax, 1.0, mapymax, mapymin, 1.0);
    grid.DrawGrid(csColor(0.2f, 0.2f, 0.2f));
    grid.DrawOrigin(csColor(0.7f, 0.0f, 0.0f));
    
    switch(viewtype)
    {
    case V2D_TOP:
      glRotatef(-90, 1.0, 0.0, 0.0);
      break;
    case V2D_FRONT:
      break;
    case V2D_SIDE:
      glRotatef(90, 0.0, 1.0, 0.0);
      break;
    }    
  }  

  if(object_view)
  {
    csRef<icssPlugin> selplug = m_csi->GetPlugin("cstudio.select.defselect");
    csRef<icssMeshSelection> meshsel = SCF_QUERY_INTERFACE(selplug, icssMeshSelection);
  
    if(meshsel->GetCount() == 1)
    {
      iMeshWrapper *mesh = meshsel->Get(0);
      drawer.ApplyTransform(mesh->GetMovable()->GetTransform().GetInverse());
    }
  }
  
  // Draw Stuff here
  csRef<icssPlugin> emplug = m_csi->GetPlugin("cstudio.plugin.emmanager");
  csRef<icssEditModeManager> emmgr = SCF_QUERY_INTERFACE(emplug, icssEditModeManager);
  csRef<icssEditMode> em = emmgr->GetActiveEditMode();
  
  if(em)
    em->OnDraw( &drawer , m_wnd);

  glColor3f(0.0, 0.0, 1.0);

  glPointSize(4);
  glBegin(GL_POINTS);
  glVertex3fv(&last3dpos.x);
  glEnd();
 
  drawer.SetLineWidth(1.0f);
  drawer.SetLineStipple(0, 0xffff);
  csColor4 c(1.0, 1.0, 1.0);
  drawer.SetColor(c);
}

void cssView2DDrawer::Move(csVector3 &dir)
{
  csVector2 move;
  switch(viewtype)
  {
  case V2D_TOP:
    move.x = dir.x;
    move.y = dir.z;
    break;
  case V2D_FRONT:
    move.x = dir.x;
    move.y = dir.y;
    break;
  case V2D_SIDE:
    move.x = dir.z;
    move.y = dir.y;
    break;
  }
  pos += move;
}

void cssView2DDrawer::Zoom(float dir)
{
  dir *= 0.5f;
  if( zoom - dir <= 0.1 )
    return;
  zoom -= dir;
}

void cssView2DDrawer::MapToView(const csVector3 &mappos, csVector2 &viewpos)
{
  if(viewtype == V2D_PERSPECTIVE)
  {
    csVector3 mappos_cam = camera3d->GetTransform().Other2This(mappos);
    
    csVector2 vp;
    camera3d->Perspective(mappos_cam, vp);
    
    viewpos.x = (int)vp.x;
    viewpos.y = -((int)vp.y) + viewarea.Height();
    
    return;
  }
  
  csVector2 tmp;
  
  switch(viewtype)
  {
    case V2D_TOP:
      tmp.x = mappos.x;
      tmp.y = mappos.z;
      break;
    case V2D_FRONT:
      tmp.x = mappos.x;
      tmp.y = mappos.y;
      break;
    case V2D_SIDE:
      tmp.x = mappos.z;
      tmp.y = mappos.y;
      break;
  }
  
  tmp -= pos;
  
  tmp.y *= -1;
  
  tmp.x *= zoom;
  tmp.y *= zoom;
  
  tmp.x += (float)viewarea.Width() / 2.0f;
  tmp.y += (float)viewarea.Height() / 2.0f;
  
  viewpos.x = (int) tmp.x;
  viewpos.y = (int) tmp.y;
}

void cssView2DDrawer::GetMapPosfromMouseCoords(int x, int y, csVector3 &omappos)
{
  csVector3 result;
  
  if(viewtype == V2D_PERSPECTIVE)
  {
    csVector2 p(x, viewarea.Height() - y);

    csVector3 tmp;
    
    camera3d->InvPerspective(p, 10, tmp);
    tmp = camera3d->GetTransform().This2Other(tmp);
    
    csVector3 origin = camera3d->GetTransform().GetOrigin();
    
    switch(ref_plane)
    {
      case CSS_PLANE_X:
        csIntersect3::SegmentXPlane(origin, tmp, 0, result);
        break;       
      case CSS_PLANE_Y:
        csIntersect3::SegmentYPlane(origin, tmp, 0, result);
        break;       
      case CSS_PLANE_Z:
        csIntersect3::SegmentZPlane(origin, tmp, 0, result);
        break;
    }       
  }
  else
  {
    int viewx = x - viewarea.xmin;
    int viewy = viewarea.Height() - y;
    
    float temp = viewarea.Width( ) / zoom;
    result.x = pos.x - temp / 2.0f + viewx * ( temp / (float)viewarea.Width( ) );
    temp = viewarea.Height( ) / zoom;
    result.y = pos.y - temp / 2.0f + viewy * ( temp / (float)viewarea.Height( ) );
  
    switch(viewtype)
    {
    case V2D_TOP:
      temp = result.y;
      result.y = 0;
      result.z = temp;
      break;
    case V2D_FRONT:
      break;
    case V2D_SIDE:
      temp = result.x;
      result.x = result.z;
      result.z = temp;
      break;
    }
  
    
  }  
  
  last3dpos = result;
  omappos = result;
}

void cssView2DDrawer::CreateBeam(int x, int y, csSegment3 &beam)
{
  csVector3 start, end;
  
  if(viewtype == V2D_PERSPECTIVE)
  {
    start = camera3d->GetTransform().GetOrigin();  
    camera3d->InvPerspective(csVector2(x, viewarea.Height() - y), 10000, end);
    end = camera3d->GetTransform().This2Other(end);
  }
  else
  {
    int viewx = x - viewarea.xmin;
    int viewy = viewarea.Height() - y;
    
    csVector2 beam2d;
    
    float temp = viewarea.Width( ) / zoom;    
    beam2d.x = pos.x - temp / 2.0f + viewx * ( temp / (float)viewarea.Width( ) );
    temp = viewarea.Height( ) / zoom;
    beam2d.y = pos.y - temp / 2.0f + viewy * ( temp / (float)viewarea.Height( ) );
    
    switch(viewtype)
    {
    case V2D_TOP:
      start.Set(beam2d.x, 10000, beam2d.y);
      end.Set(beam2d.x, -10000, beam2d.y);      
      break;
    case V2D_FRONT:
      start.Set(beam2d.x, beam2d.y, 10000);
      end.Set(beam2d.x, beam2d.y, -10000);      
      break;
    case V2D_SIDE:
      start.Set(10000, beam2d.y, beam2d.x);
      end.Set(-10000, beam2d.y, beam2d.x);      
      break;
    }
  }
  
  beam.Set(start, end);
}
  
bool cssView2DDrawer::MovePoint(const csVector3 &pos, const csVector2 &newpos2d, 
    csVector3 &newpos3d, bool y_direction)
{
  csVector2 p = newpos2d;
  csSegment3 beam;
  
  if(viewtype == V2D_PERSPECTIVE && y_direction)
  {
    csVector2 tmp;
    MapToView(pos, tmp);
    p.x = tmp.x;
  }

  // + 0.5 to automatically get it rounded
  CreateBeam((int)(p.x + 0.5f), (int)(p.y + 0.5f), beam);
      
  switch(viewtype)
  {
    case V2D_TOP:
      csIntersect3::SegmentYPlane(beam.Start(), beam.End(), pos.y, newpos3d);    
      break;
    case V2D_FRONT:
      csIntersect3::SegmentZPlane(beam.Start(), beam.End(), pos.z, newpos3d);    
      break;
    case V2D_SIDE:
      csIntersect3::SegmentXPlane(beam.Start(), beam.End(), pos.x, newpos3d);    
      break;
    case V2D_PERSPECTIVE:
      if(y_direction)
      {
        csIntersect3::SegmentZPlane(beam.Start(), beam.End(), pos.z, newpos3d);
        newpos3d.x = pos.x;
        newpos3d.z = pos.z;
      }
      else
        csIntersect3::SegmentYPlane(beam.Start(), beam.End(), pos.y, newpos3d);
      break;       
  }           
  return true;
}


/*
void cssView2DDrawer::DrawMeshSelection(icssMeshSelection *ms)
{
	int nmeshes = ms->GetCount();
	glColor3f(1.0, 0.0, 0.0);
	for(int i = 0; i < nmeshes; i++)
	{
		csRef<iMeshWrapper> mwrap =	ms->Get(i);
		DrawMesh(mwrap);
	}
}

void cssView2DDrawer::DrawThingSelection(icssThingSelection *ts)
{
  int nsel = ts->GetCount();
  glColor3f(1.0, 0.0, 0.0);
  for(int i = 0; i < nsel; i++)
  {
    csRef<icssThingPart> tp = ts->Get(i);
    DrawThingPart(tp);
  }
}

void cssView2DDrawer::DrawThingPart(icssThingPart *tp)
{
  csRef<iMeshWrapper> mesh = tp->GetMesh();
  csRef<iThingState> ts = SCF_QUERY_INTERFACE(mesh->GetMeshObject(), iThingState);
  csRef<iThingFactoryState> tfs = ts->GetFactory();

  glPushMatrix();
    ApplyObjectToCamera(mesh->GetMovable()->GetFullTransform());
    glEnableClientState(GL_VERTEX_ARRAY);
    const csVector3 *verts = tfs->GetVertices();
    glVertexPointer(3, GL_FLOAT, 0, verts);
    

    // draw points
    int vertcount = tp->GetVertexCount();
    glBegin(GL_POINTS);
    
    for(int i = 0; i < vertcount; i++)
    {
      glArrayElement(tp->GetVertexIndex(vertcount));
    }

    glEnd();
    

    // draw polygons
    int polycount = tp->GetPolygonCount();
    for(int i = 0; i < polycount; i++)
    {
      int curpoly = tp->GetPolygonIndex(i);
      int *indices = tfs->GetPolygonVertexIndices(curpoly);
      int indexcount = tfs->GetPolygonVertexCount(curpoly);

      glBegin(GL_LINE_LOOP);
        for(int j = 0;j < indexcount; j++)
        {
          glArrayElement(indices[j]);
        }

      glEnd();
    }
    glDisableClientState(GL_VERTEX_ARRAY);
    

  glPopMatrix();
}

*/
