/*
    Copyright (C) 2005 by Jan Lepper

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "cssysdef.h"
#include "csstabwindow.h"
#include "csutil/csstring.h"

#include <wx/artprov.h>


#define CSS_TW_BITMAP_SIZE 16
#define CSS_TW_BUTTON_BORDER 2
#define CSS_TW_BUTTON_SIZE (CSS_TW_BITMAP_SIZE + (CSS_TW_BUTTON_BORDER * 2))
#define CSS_TW_TABBAR_HEIGHT (tab_height + 1)
#define CSS_TW_PAGE_BORDER 3
#define CSS_TW_NC_HEIGHT (CSS_TW_TABBAR_HEIGHT + CSS_TW_PAGE_BORDER)



BEGIN_EVENT_TABLE(cssTabWindow, wxWindow)
  EVT_SIZE(cssTabWindow::OnSize)
  EVT_PAINT(cssTabWindow::OnPaint)
  EVT_LEFT_DOWN(cssTabWindow::OnLeftDown)
  EVT_ERASE_BACKGROUND(cssTabWindow::OnErase)
  EVT_BUTTON(CSS_TABWINDOW_BACK, cssTabWindow::OnButtonBack)
  EVT_BUTTON(CSS_TABWINDOW_FORWARD, cssTabWindow::OnButtonForward)
END_EVENT_TABLE()


cssTabWindow::cssTabWindow(wxWindow *parent, int id, wxPoint pos, wxSize size, int style, bool bottom) :
  wxWindow(parent, id, pos, size, style)
{
  active_page = -1;
  tabbar_y = 0;
  tab_y = 0;
  first_tab = 0;
  width = height = 0;
  show_buttons = false;
  this->bottom = bottom;  

  if(GetCharHeight() < CSS_TW_BITMAP_SIZE)
    tab_height = CSS_TW_BITMAP_SIZE + 6;
  else
    tab_height = GetCharHeight() + 6;
  
  b_back = new wxBitmapButton(this, CSS_TABWINDOW_BACK, 
    wxArtProvider::GetBitmap(wxART_GO_BACK), 
    wxPoint(0,0), wxSize(CSS_TW_BUTTON_SIZE, tab_height));
  b_back->Enable(false);

  b_forward = new wxBitmapButton(this, CSS_TABWINDOW_FORWARD, 
    wxArtProvider::GetBitmap(wxART_GO_FORWARD), 
    wxPoint(0,100), wxSize(CSS_TW_BUTTON_SIZE, tab_height));
  b_forward->Enable(false);    
  
  Layout();
}

  
wxWindow *cssTabWindow::GetPage(size_t idx)
{
  return pages[idx];
}


int cssTabWindow::GetTabWidth(int idx)
{
  int text_w, text_h;
  GetTextExtent(labels[idx], &text_w, &text_h, 0, 0);

  return text_w + 18 + CSS_TW_BITMAP_SIZE; 
}


int cssTabWindow::GetTabAt(int x, int y)
{
  if(y  >= tab_y && y < tab_y + tab_height)
  {
    if((show_buttons && x < (width - (CSS_TW_BUTTON_SIZE * 2))) || !show_buttons)
    {
      int tab_start = 0;
    
      for(size_t i = first_tab; i < pages.Length(); i++)
      {
        int tw = GetTabWidth(i);
      
        if(x >= tab_start && x < tab_start + tw)
          return i;
      
        tab_start += tw;
      }       
    }
  }
  
  return -1;
}


void cssTabWindow::SetActivePage(size_t idx)
{
  if(idx == (size_t)active_page)
    return;
  
  int old_active = active_page;
  active_page = idx;
  
  wxClientDC dc(this);
  DrawTabs(dc);
  
  LayoutPage();
  pages[idx]->Show(true);

  if(old_active >= 0)
    pages[old_active]->Show(false);
}


void cssTabWindow::DeleteAllPages()
{
  for(size_t i = 0; i < pages.Length(); i++)
    pages[i]->Destroy();
  
  first_tab = 0;
  active_page = -1;
  
  b_forward->Enable(false);
  b_back->Enable(false);
  
  
  pages.DeleteAll();
  labels.DeleteAll();
  bitmaps.DeleteAll();
  Refresh();
}


void cssTabWindow::AddPage(wxWindow *page, const char *label, const wxBitmap& bitmap)
{
  labels.Push(label);
  int idx = bitmaps.Push(bitmap);
  bitmaps[idx].SetWidth(CSS_TW_BITMAP_SIZE);
  bitmaps[idx].SetHeight(CSS_TW_BITMAP_SIZE);
  SetActivePage(pages.Push(page));
  
  UpdateButtons();
}


void cssTabWindow::DeletePage(size_t idx)
{
  wxWindow *deleted_page = pages[idx];
  
  pages.DeleteIndex(idx);
  labels.DeleteIndex(idx);
  bitmaps.DeleteIndex(idx);
  
  if(first_tab >= pages.Length())
    first_tab = pages.Length();
  
  UpdateButtons();

  if((size_t)active_page == idx)
  {
    active_page = -1;
    
    if(pages.Length())
    {
      if(idx > 0)
        SetActivePage(idx-1);
      else
        SetActivePage(0);
    }
  }

  else if((size_t)active_page > idx)
    active_page--;

  UpdateButtons();
    
  deleted_page->Show(false);
  deleted_page->Destroy();
}


void cssTabWindow::ReplacePage(size_t idx, wxWindow* new_page, const char* label, const wxBitmap &bitmap)
{

  wxWindow *old_page = pages[idx];
  
  labels.Put(idx, label);

  bitmaps.Put(idx, bitmap);
  bitmaps[idx].SetWidth(CSS_TW_BITMAP_SIZE);
  bitmaps[idx].SetHeight(CSS_TW_BITMAP_SIZE);
  
  pages.Put(idx, new_page);
  
  wxClientDC dc(this);
  DrawTabs(dc);

  UpdateButtons();
    
  bool active = (idx == (size_t)active_page);
  
  if(active)
    LayoutPage();
  
  new_page->Show(active);  
  
  old_page->Show(false);
  old_page->Destroy();  

}


void cssTabWindow::DeleteOtherPages(size_t idx)
{
  wxWindow* page = pages[idx];
  csString label = labels[idx];
  wxBitmap bitmap = bitmaps[idx];
  
  for(size_t i = 0; i < pages.Length(); i++)
  {
    if((size_t)active_page != idx)
      pages[i]->Destroy();
  }
  
  pages.DeleteAll();
  labels.DeleteAll();
  bitmaps.DeleteAll();
  
  pages.Push(page);
  labels.Push(label);
  bitmaps.Push(bitmap);
  
  active_page = -1;
  SetActivePage(0);
}


bool cssTabWindow::Layout()
{
  GetClientSize(&width, &height);
  
  tabbar_y = bottom ? height - CSS_TW_TABBAR_HEIGHT : 0;   
  tab_y = (bottom ? tabbar_y : tabbar_y + (CSS_TW_TABBAR_HEIGHT - tab_height));
  
  b_back->Move(width - (CSS_TW_BUTTON_SIZE * 2), tab_y);
  b_forward->Move(width - CSS_TW_BUTTON_SIZE, tab_y);

  LayoutPage();

  UpdateButtons();  
  
  return true;
}


void cssTabWindow::LayoutPage()
{
  if(active_page >= 0)
  {
    if(pages.Length() != 1)
    {
      int page_height = height - CSS_TW_NC_HEIGHT;
      int page_y = bottom ? 0 : CSS_TW_NC_HEIGHT;
      pages[active_page]->SetSize(0, page_y, width , page_height);  
    }
    else
      pages[active_page]->SetSize(0, 0, width , height);  
  }
}


void cssTabWindow::UpdateButtons()
{
  int tabs_length = 0;
  for(size_t idx = first_tab; idx < pages.Length(); idx++)
    tabs_length += GetTabWidth(idx);
  
  bool tabs_fit = tabs_length < width;  
  
  b_back->Enable(first_tab);
  b_forward->Enable(!tabs_fit && (first_tab < pages.Length() - 1));
  
  show_buttons = first_tab || !tabs_fit;
  
  b_back->Show(show_buttons);
  b_forward->Show(show_buttons);
}


void cssTabWindow::DrawTabs(wxDC &dc)
{
  if(pages.Length() <= 1)
    return;

  dc.BeginDrawing();
    
  wxPen pen_highlight (wxSystemSettings::GetColour(wxSYS_COLOUR_3DHIGHLIGHT), 1, wxSOLID);
  wxPen pen_shadow (wxSystemSettings::GetColour(wxSYS_COLOUR_3DSHADOW), 1, wxSOLID);

  dc.SetFont(GetFont());
  
  int active_start = 0;
  int active_width = 0;

  dc.Clear();

  // tabs
  
  if(show_buttons)
    dc.SetClippingRegion(0, tab_y, width - (CSS_TW_BUTTON_SIZE * 2), tab_height);      
  
  int tab_x = 0;
  
  for(size_t idx = first_tab; idx < pages.Length(); idx++)
  {
    int tab_width = GetTabWidth(idx);

    bool active = (idx == (size_t)active_page);    
    
    if(active)
    {
      active_start = tab_x;
      active_width = tab_width;
      
      //right/bottom border
      dc.SetPen(pen_shadow);
      dc.DrawLine(tab_x + tab_width - 1, tab_y, tab_x + tab_width - 1, tab_y + tab_height);
      if(bottom)
        dc.DrawLine(tab_x, tab_y + tab_height - 1, tab_x + tab_width, tab_y + tab_height - 1); 
    
      //left/top border
      dc.SetPen(pen_highlight);  
      dc.DrawLine(tab_x, tab_y, tab_x, tab_y + tab_height);
      if(!bottom)
        dc.DrawLine(tab_x, tab_y, tab_x + tab_width, tab_y);
    }    
    else if(idx != 0  && idx != (size_t)active_page + 1)
    {
      //separator
      dc.SetPen(pen_shadow);        
      dc.DrawLine(tab_x, tab_y + 2, tab_x, tab_y  + tab_height - 2);
    }
    
    //label
    int label_x = active ? tab_x + 5 : tab_x + 4;
    int label_y = active ? tab_y + 3 : tab_y + 4;
    dc.DrawText(labels[idx], label_x + CSS_TW_BITMAP_SIZE + 2, label_y);  
    dc.DrawBitmap(bitmaps[idx], label_x, label_y, true);
    
    tab_x += tab_width;
  }
  
  dc.DestroyClippingRegion();
  
  // page border  

  if(bottom)
    dc.SetPen(pen_shadow);
  else
    dc.SetPen(pen_highlight);  
  
  int border_y = bottom ? tabbar_y - 1: CSS_TW_TABBAR_HEIGHT;

  dc.DrawLine(0, border_y, active_start + 1, border_y);  
  dc.DrawLine(active_start + active_width - 1, border_y, width, border_y);  
  if(show_buttons)
    dc.DrawLine(width - (CSS_TW_BUTTON_SIZE * 2), border_y, width, border_y);  
  
  dc.EndDrawing();
}


void cssTabWindow::OnErase(wxEraseEvent &event)
{
  if(!pages.Length())
    event.Skip();
}


void cssTabWindow::OnPaint(wxPaintEvent &event)
{
  wxPaintDC dc (this);
  DrawTabs(dc);
}


void cssTabWindow::OnLeftDown(wxMouseEvent &event)
{
  int tab = GetTabAt(event.GetX(), event.GetY());
  if(tab >= 0)
    SetActivePage(tab);
}


void cssTabWindow::OnButtonBack(wxCommandEvent &event)
{
  first_tab--;
  wxClientDC dc(this);
  DrawTabs(dc);
  UpdateButtons();  
}


void cssTabWindow::OnButtonForward(wxCommandEvent &event)
{
  first_tab++;
  wxClientDC dc(this);
  DrawTabs(dc);
  UpdateButtons();  
}
