/*
    Copyright (C) 2003 by Philipp Aumayr

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "cssysdef.h"
#include "cst-config.h"
#include "csutil/scf.h"
#include "iutil/vfs.h"
#include "csutil/cscolor.h"
#include "cstool/csview.h"
#include "cstool/initapp.h"
#include "csgeom/csrect.h"
#include "iutil/eventq.h"
#include "iutil/event.h"
#include "iutil/objreg.h"
#include "iutil/csinput.h"
#include "iutil/virtclk.h"
#include "iengine/sector.h"
#include "iengine/engine.h"
#include "iengine/camera.h"
#include "iengine/light.h"
#include "iengine/texture.h"
#include "iengine/mesh.h"
#include "iengine/movable.h"
#include "iengine/material.h"
#include "imesh/sprite3d.h"
#include "imesh/object.h"
#include "ivideo/graph3d.h"
#include "ivideo/graph2d.h"
#include "ivideo/txtmgr.h"
#include "ivideo/texture.h"
#include "ivideo/material.h"
#include "ivideo/fontserv.h"
#include "igraphic/imageio.h"
#include "imap/loader.h"
#include "imap/writer.h"
#include "ivaria/reporter.h"
#include "ivaria/stdrep.h"
#include "csutil/cmdhelp.h"
#include "csutil/event.h"
#include "iengine/campos.h"
#include "ivaria/engseq.h"
#include "csutil/stringarray.h"
#include "iutil/cfgmgr.h"
#include "ivideo/wxwin.h"
#include "iutil/stringarray.h"

#ifdef WIN32
// WIN32 specific
#include "csutil/win32/win32.h"
#endif

#include "plugins/icsscommandmanager.h"
#include "plugins/icsstextoutput.h"

#include "cssProgressMeter.h"

#include <wx/gauge.h>
#include <wx/confbase.h>

#include "cstudiocs.h"

#include "cssmainframe.h"
#include "cssview3dframe.h"

#ifndef CS_BUILD_SHARED_LIBS
CS_IMPLEMENT_FOREIGN_DLL
#endif

SCF_IMPLEMENT_IBASE(CStudioCS)
SCF_IMPLEMENTS_INTERFACE(iCStudioCS)
SCF_IMPLEMENTS_EMBEDDED_INTERFACE(icssContentCallback)
SCF_IMPLEMENTS_EMBEDDED_INTERFACE(icssActiveSectorCallback)
SCF_IMPLEMENT_IBASE_END

SCF_IMPLEMENT_EMBEDDED_IBASE(CStudioCS::cssContentCallback)
SCF_IMPLEMENTS_INTERFACE(icssContentCallback)
SCF_IMPLEMENT_EMBEDDED_IBASE_END

SCF_IMPLEMENT_EMBEDDED_IBASE(CStudioCS::cssActiveSectorCallback)
SCF_IMPLEMENTS_INTERFACE(icssActiveSectorCallback)
SCF_IMPLEMENT_EMBEDDED_IBASE_END

// global pointer for static event handler
CStudioCS *cstudiocs;

CStudioCS::CStudioCS(iObjectRegistry *obj_reg)
{
  SCF_CONSTRUCT_IBASE(0);
  SCF_CONSTRUCT_EMBEDDED_IBASE(scficssContentCallback);
  SCF_CONSTRUCT_EMBEDDED_IBASE(scficssActiveSectorCallback);

  cstudiocs = this;
  do_freelook = false;
  initialized = false;
  object_reg = obj_reg;
  redraw_views = false;
}

CStudioCS::~CStudioCS()
{
  SCF_DESTRUCT_EMBEDDED_IBASE(scficssActiveSectorCallback);
  SCF_DESTRUCT_EMBEDDED_IBASE(scficssContentCallback);
  SCF_DESTRUCT_IBASE();
  cstudiocs = 0;
}

void CStudioCS::SetupFrame ()
{
  /*
    Make CS' GL context current - our own GL stuff has prolly changed the
    current context, so set it to CS' so it can draw correctly.
   */
  if (mainfrm->view3d)
  if(!g3d->GetDriver2D()->PerformExtension ("setglcontext"))
    printf("error setting gl context!!!! ARGL!");

  // First get elapsed time from the virtual clock.
  csTicks elapsed_time = vc->GetElapsedTicks ();

  // Now rotate the camera according to keyboard state
  float speed = (elapsed_time / 1000.0) * (0.03 * 30);

  if(view)
  {
    int w, h; 
    view3d->GetClientSize(&w, &h);
    view->SetRectangle(0, 0,w, h);

    iCamera* c = view->GetCamera();
    c->SetPerspectiveCenter(w/2, h/2);
    c->SetFOVAngle(90, w);
    
    if (kbd->GetKeyState (CSKEY_RIGHT))
      c->GetTransform ().RotateThis (CS_VEC_ROT_RIGHT, speed);
    if (kbd->GetKeyState (CSKEY_LEFT))
      c->GetTransform ().RotateThis (CS_VEC_ROT_LEFT, speed);
    if (kbd->GetKeyState (CSKEY_PGUP))
      c->GetTransform ().RotateThis (CS_VEC_TILT_UP, speed);
    if (kbd->GetKeyState (CSKEY_PGDN))
      c->GetTransform ().RotateThis (CS_VEC_TILT_DOWN, speed);
    if (kbd->GetKeyState (CSKEY_UP))
      c->Move (CS_VEC_FORWARD * 6 * speed);
    if (kbd->GetKeyState (CSKEY_DOWN))
      c->Move (CS_VEC_BACKWARD * 6 * speed);

    // Tell 3D driver we're going to display 3D things.
    if (!g3d->BeginDraw (engine->GetBeginDrawFlags () | CSDRAW_3DGRAPHICS))
      return;

    g3d->GetDriver2D()->Clear(0);
    view->Draw ();
  }
  // Tell the camera to render into the frame buffer.
  
}

void CStudioCS::FinishFrame ()
{
  g3d->FinishDraw ();
  g3d->Print (0);
}

bool CStudioCS::HandleEvent (iEvent& ev)
{
  if (ev.Name == csevProcess (object_reg))
  {
    SetupFrame ();
    return true;
  }
  else if (ev.Name == csevFinalProcess (object_reg))
  {
    FinishFrame ();
    return true;
  }
  else if (ev.Name == csevQuit (object_reg))
  {
    return true;
  }
  else if ((ev.Name == csevKeyboardDown (object_reg)) && 
    (csKeyEventHelper::GetCookedCode (&ev) == CSKEY_ESC))
  {
    csRef<iEventQueue> q (CS_QUERY_REGISTRY (object_reg, iEventQueue));
    if (q) q->GetEventOutlet()->Broadcast (csevQuit (object_reg));
    return true;
  }
  else if (ev.Name == csevMouseDown (object_reg, 0))
  {
    if (csMouseEventHelper::GetButton(&ev))
    {
      do_freelook = true;
      int width, height;
      view3d->GetClientSize(&width, &height);
      view3d->WarpPointer (width / 2, height / 2);
      
      view3d->SetCursor(wxCursor(wxCURSOR_BLANK));
    }
  }
  else if (ev.Name == csevMouseUp (object_reg, 0))
  {
    if (csMouseEventHelper::GetButton(&ev) == 2)
    {
      do_freelook = false;

      view3d->SetCursor(wxCursor(wxCURSOR_ARROW));
    }
  }
  else if (ev.Name == csevMouseMove (object_reg, 0))
  {
    if (view)
    {
      static bool first_time = false;
      
      if (do_freelook)
      {
        if (!first_time)
        {
          int last_x, last_y;
          
          last_x = csMouseEventHelper::GetX(&ev);
          last_y = csMouseEventHelper::GetY(&ev);
          
          int width, height;
          view3d->GetClientSize(&width, &height);
          view3d->WarpPointer (width / 2, height / 2);
          
          // Calculations use CS coordinates!
          int widthc, heightc;
          widthc = g2d->GetWidth();
          heightc = g2d->GetHeight();
          
          bool inverse_mouse = false;
          
          float x = (float) -((float)(last_y - (heightc / 2) )) / (heightc*2)*(1-2*(int)inverse_mouse);
          float y = (float) ((float)(last_x - (widthc / 2) )) / (widthc*2);
          
          csMatrix3 mat = view->GetCamera ()->GetTransform ().GetO2T ();
          if(x)
            mat = csXRotMatrix3(x) * mat;
          if(y)
            mat *= csYRotMatrix3(y);
          view->GetCamera ()->SetTransform ( csOrthoTransform 
              (mat, view->GetCamera ()->GetTransform ().GetOrigin ()));
        }
        else
          first_time = false;
      }
      else
        first_time = true;
    }
  }

  return false;
}

bool CStudioCS::StaticEventHandler (iEvent& ev)
{
  return cstudiocs->HandleEvent (ev);
}

bool CStudioCS::update()
{
  if(initialized)
  {
    if (vc)
        vc->Advance();
    ev_queue->Process();
    
    if(redraw_views)
    {
      size_t cbcount = m_redrawviewcb.Length();
      for(size_t i = 0; i < cbcount; i++)
        m_redrawviewcb.Get(i)->Redraw();
      redraw_views = false;
    }
  }
  return true;
}

bool CStudioCS::Init(wxWindow *parent, const char *title, 
                     int resx, int resy,  bool useopengl)
{
  vfs = csInitializer::SetupVFS (object_reg);
  if (vfs)
  {
    vfs->Mount ("/cstudio", "$*data$/, $*cstudio.zip");
  }

  csRef<iConfigManager> confmgr = CS_QUERY_REGISTRY(object_reg, iConfigManager);
  confmgr->SetDynamicDomainPriority(iConfigManager::PriorityMax);
  confmgr->GetDynamicDomain()->SetStr("Video.OpenGL.Canvas", "crystalspace.graphics2d.wxgl");

/*  if (!csInitializer::SetupConfigManager (object_reg, "/cstudio/config/cstudio.cfg",
    "CrystalStudio"))
  {
      csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
    	"crystalspace.application.cstudio",
	 "Can't setup config manager!");
      return false;
  }
*/
  if (!csInitializer::RequestPlugins (object_reg,
    CS_REQUEST_VFS,
    CS_REQUEST_OPENGL3D,
    CS_REQUEST_ENGINE,
    CS_REQUEST_FONTSERVER,
    CS_REQUEST_IMAGELOADER,
    CS_REQUEST_LEVELLOADER,
    CS_REQUEST_LEVELSAVER,
    CS_REQUEST_REPORTER,
    CS_REQUEST_REPORTERLISTENER,
    CS_REQUEST_PLUGIN("crystalspace.utilities.sequence.engine", iEngineSequenceManager),
    CS_REQUEST_END))
  {
      csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
    	  "crystalspace.application.cstudio",
	  "Can't initialize plugins!");
      return false;
  }

  if (!csInitializer::SetupEventHandler (object_reg, StaticEventHandler))
  {
    csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
    	"crystalspace.application.cstudio",
	"Can't initialize event handler!");
    return false;
  }

  // Check for commandline help.
  if (csCommandLineHelper::CheckHelp (object_reg))
  {
    csCommandLineHelper::Help (object_reg);
    return false;
  }

  // The virtual clock.
  vc = CS_QUERY_REGISTRY (object_reg, iVirtualClock);
  if (vc == 0)
  {
    csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
    	"crystalspace.application.cstudio",
	"Can't find the virtual clock!");
    return false;
  }

  // Find the pointer to engine plugin
  engine = CS_QUERY_REGISTRY (object_reg, iEngine);
  if (engine == 0)
  {
    csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
    	"crystalspace.application.cstudio",
	"No iEngine plugin!");
    return false;
  }
  engine->SetSaveableFlag (true);
  
  loader = CS_QUERY_REGISTRY (object_reg, iLoader);
  if (loader == 0)
  {
    csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
    	"crystalspace.application.cstudio",
    	"No iLoader plugin!");
    return false;
  }

  saver = CS_QUERY_REGISTRY (object_reg, iSaver);
  if (loader == 0)
  {
    csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
    	"crystalspace.application.cstudio",
    	"No iLoader plugin!");
    return false;
  }

  g3d = CS_QUERY_REGISTRY (object_reg, iGraphics3D);
  if (g3d == 0)
  {
    csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
    	"crystalspace.application.cstudio",
    	"No iGraphics3D plugin!");
    return false;
  }

  g2d = g3d->GetDriver2D();
  if(g2d == 0)
  {
    csReport(object_reg, CS_REPORTER_SEVERITY_ERROR,
      "crystalspace.application.cstudio",
      "No iGraphics2D plugin!");
    return false;
  }

  kbd = CS_QUERY_REGISTRY (object_reg, iKeyboardDriver);
  if (kbd == 0)
  {
    csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
    	"crystalspace.application.cstudio",
    	"No iKeyboardDriver plugin!");
    return false;
  }

  engseqmgr = CS_QUERY_REGISTRY (object_reg, iEngineSequenceManager);
  if(engseqmgr == 0)
  {
    csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
      "crystalspace.application.cstudio",
      "No iEngineSequenceManager plugin!");
    return false;
  }

  view3d = new cssView3DFrame(parent);

  view3d->Show(true);

  csRef<iWxWindow> wxwin = SCF_QUERY_INTERFACE(g2d, iWxWindow);
  if(wxwin)
    wxwin->SetParent(view3d);



  // Open the main system. This will open all the previously loaded plug-ins.
  if (!csInitializer::OpenApplication (object_reg))
  {
    csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
    	"crystalspace.application.cstudio",
    	"Error opening system!");
    return false;
  }

/*
  view = csPtr<iView> (new csView (engine, g3d));
  iGraphics2D* g2d = g3d->GetDriver2D ();
  view->SetRectangle (0, 0, g2d->GetWidth (), g2d->GetHeight ());
*/



  ev_queue = CS_QUERY_REGISTRY(object_reg, iEventQueue);
  if (!ev_queue)
    return false;

  /// Load Console stuff
  RegisterConsole();
	
#ifdef WIN32
  //// win32 specific!
  win32ass = CS_QUERY_REGISTRY(object_reg, iWin32Assistant);
  if(!win32ass)
     return false;
  // make sure CS doesn't grab messages, but feed it directly  
  win32ass->UseOwnMessageLoop(false);
  //// end win32 specific
#endif
  LoadPlugins("cstudio.select.defselect");
  LoadPlugins("cstudio.plugin.");

  LoadStandardMaterials();
  
  initialized = true;
  
  return true;
}

int CStudioCS::shutdown()
{

  /* this causes a crash for some unknown reason, and it seems not to be necessary because
      cscmdQuit is only used by cs's default messageloop which we dont't use
  if (ev_queue)
  {
    csRef<iEvent> event = ev_queue->CreateEvent(cscmdQuit);
    ev_queue->Post(event);
    ev_queue->Process();
  }*/
  
  csInitializer::CloseApplication(object_reg);

  m_plugins.DeleteAll();
  
  initialized = false;

  return 0;
}

void CStudioCS::UnloadMap()
{
  engine->DeleteAll();

  engseqmgr->RemoveSequences();
  engseqmgr->RemoveTriggers();
}

bool CStudioCS::LoadMap(char *path, char *file, wxGauge *gauge)
{
  if(!vfs->Exists( path ) )
    return false;

  UnloadMap();

  vfs->ChDir(path);

  if( !loader->LoadMapFile( file, true, 0, false, false ))
  {
    csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
      "crystalspace.application.cstudio",
      "Couldn't load level!");
    return false;
  }
		
  csRef<cssProgressMeter> pMeter = new cssProgressMeter(100, gauge);

  engine->Prepare(pMeter);

  csVector3 pos (0, 0, 0);
  if (engine->GetCameraPositions ()->GetCount () > 0)
  {
    // There is a valid starting position defined in the level file.
    iCameraPosition* campos = engine->GetCameraPositions ()->Get (0);
    active_sector = engine->GetSectors ()->FindByName (campos->GetSector ());
    pos = campos->GetPosition ();
  }
  else
  {
    // We didn't find a valid starting position. So we default
    // to going to room called 'room' at position (0,0,0).
    		active_sector = engine->GetSectors ()->FindByName ("room");
  }
  if (!active_sector)
  {
    csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
    	"crystalspace.application.cstudio",
      	"Can't find a valid starting position!");
    return false;
  }
  if(!view)
    view = new csView(engine, g3d);
  view->GetCamera ()->SetSector (active_sector);
  view->GetCamera ()->GetTransform ().SetOrigin (pos);


  ///// @@@@@@@@@@@@@@@@@@
  ///// we don't need animation for now
  ///// later on well have to add some support for that
  ///// @@@@@@@@@@@@@@@@@
  engseqmgr->RemoveTriggers();
  engseqmgr->RemoveSequences();

  OnMaterialListChanged();
  OnMeshListChanged();
  OnSectorListChanged();

  return true;  
}

bool CStudioCS::SaveMap(char *path, char* file)
{
  if(!vfs->Exists( path ) )
    return false;

  vfs->ChDir(path);

  if (!saver->SaveMapFile( file ) )
    return false;

  return true;
}

iObjectRegistry *CStudioCS::getObjectRegistry()
{ 
  return object_reg; 
}

iSector* CStudioCS::getActiveSector()
{
  return active_sector;
}

void CStudioCS::SetActiveSector(iSector *sector)
{
  active_sector = sector;
  OnActiveSectorChanged();
  if(!view)
    view = new csView(engine, g3d);
  view->GetCamera()->SetSector(sector);
}

void CStudioCS::RegisterActiveSectorCB(icssActiveSectorCallback *cb)
{
  m_activesectorcb.Push(cb);
}

void CStudioCS::UnregisterAcitveSectorCB(icssActiveSectorCallback *cb)
{
  m_activesectorcb.Delete(cb);
}

void CStudioCS::RegisterRedrawViewCB(icssRedrawViewCallback *cb)
{
  m_redrawviewcb.Push(cb);
}

void CStudioCS::UnregisterRedrawViewCB(icssRedrawViewCallback *cb)
{
  m_redrawviewcb.Delete(cb);
}

void CStudioCS::RedrawViews()
{
  redraw_views = true;
}

void CStudioCS::RegisterSelection()
{
  csRef<icssPlugin> selectionplug = GetPlugin("cstudio.select.defselect");
}

void CStudioCS::RegisterConsole()
{
  csRef<icssPlugin> consoleplug = GetPlugin("cstudio.cmd.console");
  if(consoleplug)
    textout = SCF_QUERY_INTERFACE(consoleplug, icssTextOutput);
}

size_t CStudioCS::LoadPlugins(const char *domain)
{
  size_t curplugc = m_plugins.Length();

  csRef<iStringArray> strarr = iSCF::SCF->QueryClassList(domain);
  size_t nplugins = strarr->Length();
  for(size_t i = 0; i < nplugins; i++)
  {
    LoadPlugin(strarr->Get(i));
  }
  return m_plugins.Length() - curplugc; // return number of loaded plugins
}

csRef<icssPlugin> CStudioCS::LoadPlugin(const char *name)
{
  csRef<icssPlugin> plug = SCF_CREATE_INSTANCE(name, icssPlugin);
  if(!plug)
	{
    csString err = "couldn't create instance of ";
    err.Append(name);
    err.Append("\n");
    		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
    "cstudio.application", err);
    return 0;
  }

  if(!plug->Initialize((iCStudioCS*)this))
  {
    csString err = "couldn't initialize instance of ";
    err.Append(name);
    err.Append("\n");
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
    "cstudio.application", err);
    return 0;
  }

  m_plugins.PushSmart(plug);

  return plug;
}

csRef<icssPlugin> CStudioCS::GetPlugin(const char *name)
{
  size_t count = m_plugins.Length();
  for(size_t i = 0; i < count; i++)
  {
    csRef<icssPlugin> p = m_plugins.Get(i);
    if(!p)
    {
      csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
      "cstudio.application",
      "Load Plugin inserted a 0 pointer as a loaded plugin");
      continue;
    }

    if(strcmp(p->GetName(), name) == 0)
    return csPtr<icssPlugin>(p);
  }

  csRef<icssPlugin> p = LoadPlugin(name);
  return p;
}

size_t CStudioCS::GetPluginCount()
{
  return m_plugins.Length();
}

csRef<icssPlugin> CStudioCS::GetPlugin(unsigned int idx)
{
  if(idx < m_plugins.Length())
    return m_plugins.Get(idx);

  return 0;
}

void CStudioCS::RegisterContentCB(icssContentCallback *cb)
{
  m_contentcb.PushSmart(cb);
}

void CStudioCS::UnregisterContentCB(icssContentCallback *cb)
{
  m_contentcb.Delete(cb);
}

/// icssContentCallback implementation
void CStudioCS::OnMeshListChanged()
{
  size_t n = m_contentcb.Length();
  for(size_t i = 0; i < n; i++)
    m_contentcb.Get(i)->OnMeshListChanged();
  RedrawViews();
}

void CStudioCS::OnSectorListChanged()
{
  size_t n = m_contentcb.Length();
  for(size_t i = 0; i < n; i++)
    m_contentcb.Get(i)->OnSectorListChanged();
  RedrawViews();    
}

void CStudioCS::OnMaterialListChanged()
{
  size_t n = m_contentcb.Length();
  for(size_t i = 0; i < n; i++)
    m_contentcb.Get(i)->OnMaterialListChanged();
  RedrawViews();    
}

/// icssActiveSectorCallback implementation
void CStudioCS::OnActiveSectorChanged()
{
  size_t n = m_activesectorcb.Length();
  for(size_t i = 0; i < n; i++)
    m_activesectorcb.Get(i)->OnActiveSectorChanged();
  RedrawViews();    
}

void CStudioCS::Print(const char *msg)
{
  if(textout)
    textout->Print(msg);
}

void CStudioCS::Printf(const char *format, ...)
{
  va_list args;
  va_start (args, format);
  
  csString temp;
  temp.FormatV (format, args);
  
  Print(temp);
  
  va_end (args);
}

iMaterialWrapper* CStudioCS::GetActiveMaterial()
{ 
  return m_activemat;
}

void CStudioCS::SetActiveMaterial(iMaterialWrapper *mat)
{
  m_activemat = mat;
}

void CStudioCS::LoadStandardMaterials()
{
   iTextureManager* txtmgr = g3d->GetTextureManager ();
   iTextureWrapper* txt = loader->LoadTexture ("stdtex",
        "/lib/std/cslog2.png", CS_TEXTURE_3D, txtmgr, true);
   if (txt == 0)
   {
     csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
               "crystalspace.application.simple",
               "Error loading texture!");
                return;
   }
    csRef<iMaterialWrapper> mat = engine->FindMaterial("stdtex");
    if(!mat)
      return;

    SetActiveMaterial(mat);
}
